##############################################################################
#
# Copyright (c) 2004 Alexandre Fernandez (alex@grotte.org). All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################
from elementtree.ElementTree import fromstring, tostring
from elementtree.ElementTree import Element, SubElement
from Avatar import Avatar
from World import World
from HappyChatSessionManager import HappyChatZopeSessionManager
import time, random, re, string
from zLOG import LOG, INFO, DEBUG
from twisted.internet.protocol import Protocol

"""
Server definition and server managment
"""

class UnReconizedCommand(Exception):
    """Exception called when a unreconized command was send to the server"""
    pass #XXX: have to take a look at exception management.


class HappyChatDefaultProtocol(Protocol):
    """
    Default protocol for HappyChat Server.
    only 6 commands :
      pong : request a session on the server
      speak : say somethink is the curent room
      nickname : change nickname (not in use)
      roominfo : ask room information (list avatar)
      avatarinfo: ask information about an avatar (list properties)
      quit: disconnect the client properly

    Each command is binded to a method of this class prefixed by 'do_'
    par exemple : speak command is bind to do_speak method
    
    """
    
    def getSocket(self):
        """Return the socket of the curent connection
        @return: Socket
        @rtype: L{socket}
        """
        return self.transport.socket
    
    def connectionMade(self):
        """Call when a new connection is made
        add a pending information to the world
        """
        LOG('HappyChat', INFO, "Client connection")
        addr = ''
        socket = self.getSocket()
        world = World()
        world.addPending(socket, addr)

    def connectionLost(self, reason):
        """Call when a connection is close
        delete avatar and clean structures
        """
        if hasattr(self,'avatar'):
            avatar = self.avatar
            world = World()
            world.killByConnection(self)
            LOG('HappyChat', INFO, "Client disconnection")

    def dataReceived(self, data):
        """Call on each query
        get the command to execute in the raw XML structure
        the executed action can me a method of this class
        or a function in a module in the actions package
        
        @return: None
        
        @param data: raw XML structure take from socket read
        @type data: C{string}
        """
        server = self.factory

        server.startExternalAccess()
        
        try:
            socket = self.getSocket()
            data = string.replace(data,'\x00', '') 
            command, element = self.getCommand(data)           
            command(socket, element)
        except UnReconizedCommand:
            LOG('HappyChat', DEBUG, "Parse error on " + data)


        server.stopExternalAccess()
        
    def getCommand(self, data):
        """
        Get the command of the raw XML structure
        
        @return the command to do.
        @rtype: C{tuple} of (C{method}, L{Element}).
        
        @param data: XML raw string structure.
        @type data: C{string} from a socket read.

        """
        command = None
        element = None
        try:
            LOG('HappyChat', DEBUG,'parsing ' + data)
            element = fromstring(data)
            command_id = 'do_' + element.tag 
            command = getattr(self, command_id, None)
            LOG('HappyChat' , DEBUG, 'get command ' + str(command))
        except:
            LOG('HappyChat', DEBUG, 'Command parsing error on ' + data)

        if command is None:
            raise UnReconizedCommand
            
        return command, element
    

    def getAvatar(self, socket):
        """
        Get the default avatar by given socket

        @return: Avatar
        @rtype: L{Avatar}

        @param socket: socket where data was rode
        @type socket: L{socket}
        """
        world = World()
        avatar = world.getAvatarBySocket(socket)
        return avatar
        
    def do_speak(self, socket, element):
        """
        handle a speak request from avatar over TCP socket
        @return: None

        @param avatar: Avatar who ask the speak query
        @type vatar: L{Avatar}

        @param element: The element created by parsing the received XML
        @type element: L{Element}
        """        
        avatar = self.getAvatar(socket)
        text = element.text
        speak = Element('speak')
        speak.set('from', avatar.getNickname())
        speak.text = text
        speak_xml = tostring(speak)
        avatar.sendToRoom(speak_xml)
        del speak_xml
        del speak

    def do_roominfo(self, socket, element):
        """
        handle a room information request from avatar over TCP socket
        
        @return: None

        @param avatar: Avatar who ask the speak query
        @type vatar: L{Avatar}

        @param element: The element created by parsing the received XML
        @type element: L{Element}
        """
        avatar = self.getAvatar(socket)
        myRoom = avatar.getRoom()
        room = Element('room')
        room.set('id', myRoom.getId())
        for content in myRoom.getAvatars():
            nickname = Element('nickname')
            nickname.text = content.getNickname()
            room.append(nickname)
        
        room_xml = tostring(room)
        avatar.send(room_xml)
        del room_xml
        del room
    

    def do_nickname(self, socket, element):
        """
        handle a nickname change ask by avatar over TCP socket
        @return: None
        
        @param avatar: Avatar who ask the speak query
        @type vatar: L{Avatar}
        
        @param element: The element created by parsing the received XML
        @type element: L{Element}
        """
        avatar = self.getAvatar(socket)
        myWorld = avatar.getRoom().getWorld()
        nickname = element.text
        if not myWorld.nicknameExists(nickname):
            avatar.setNickname(nickname)
            response = Element('nickname')
            response.text = nickname
        else:
            response = Element('error')
            response.set('type', 'NICKNAME')
            response.text = "Nickname exists please choose an other one"
                  
        response_xml = tostring(response)
        avatar.send(response_xml)
        del response_xml
        del response
        
    def do_avatarinfo(self, socket, element):
        """
        handle a avatar information from avatar over TCP socket.
        @return: None
        
        @param avatar: Avatar who ask the speak query
        @type vatar: L{Avatar}
        
        @param element: The element created by parsing the received XML
        @type element: L{Element}
        """
        avatar = self.getAvatar(socket)
        nickname = element.text
        myRoom = avatar.getRoom()
        
        target = myRoom.getAvatarByNickname(nickname)
        if target is not None:
            info_xml = target.getInfoXML()
            avatar.send(info_xml)
            del info_xml
            del target
    

    def do_quit(self, socket, element):
        """
        handle a quit request, disconnect the avatar, close the socket
        and remove avatar from memory
        
        @return: None

        @param avatar: Avatar who ask the speak query
        @type vatar: L{Avatar}

        @param element: The element created by parsing the received XML
        @type element: L{Element}
        """
        avatar = self.getAvatar(socket)
        myRoom = avatar.getRoom()
        myWorld = myRoom.getWorld()
        quit = element
        quit.set('from', avatar.getNickname())
        quit_xml = tostring(quit)
        avatar.sendToRoom(quit_xml)
        myWorld.kill(avatar.getSocket())
            
    def do_pong(self, socket, element):
        """
        handle a pong request this is the first request ask by avatar
        client send a session_id.
        
        @return: None
          
        @param avatar: Avatar who ask the speak query
        @type vatar: L{Avatar}

        @param element: The element created by parsing the received XML
        @type element: L{Element}
        
        """
        world = World()

        session_id = element.attrib.get('session_id')
        SessionManager = HappyChatZopeSessionManager()
        session = SessionManager.getSession(session_id)
        nickname = session.getNickName()
        address = session.getAddress()
        pending_address = world.getPendingInfos(socket)

        newAvatar = Avatar(socket, nickname, address=pending_address)
        self.avatar = newAvatar
        SessionManager.delSession(session_id)
        
        mainRoom = world.getMainRoom()
        world.born(newAvatar, mainRoom, self)
        world.delPending(socket)
        
        welcome = Element('welcome')
        welcome_xml = tostring(welcome)
        newAvatar.send(welcome_xml)
        del welcome_xml
        del welcome
    

