##############################################################################
#
# Copyright (c) 2004 Alexandre Fernandez (alex@grotte.org). All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################
from zLOG import LOG, INFO, DEBUG
import HappyChatProtocol
from threading import Thread, Lock
from twisted.internet import protocol
from twisted.internet import reactor
from World import World
from Room import Room
from HappyChatSessionManager import HappyChatZopeSessionManager

def HappyServer(id='HappyChat', hostname='localhost', port=1234,
                protocol=HappyChatProtocol.HappyChatDefaultProtocol):
    """
    Return the _Happyserver singleton instance

    @return: The unique instance of the _HappyServer class
    @rtype: L{_HappyServer}
    
    """
    if _HappyServer.instance == None:
        return _HappyServer(id, hostname, port, protocol)
    else:
        return _HappyServer.instance

class _HappyServer(protocol.ServerFactory):


    """Server Singleton
    
    can be call by the HappyServer method
    @cvar instance: Unique instance of the singleton
    """
    
    instance = None
    protocol = None
    
    def __init__(self, id, hostname, port, protocol):
        """
        @param protocol: A protocol class that define handler method.          
        @type protocol: L{HappyChatProtocol}
        """
        self.lock = Lock()
        _HappyServer.instance = self
        _HappyServer.protocol = protocol
        self.runing = False
        self.hostname = hostname
        self.port = port
        self.id = id

    def startFactory(self):
        """Construct the server structure
        """
        world = World()
        mainRoom = Room('main', world)

    def stopFactory(self):
        """Delete server structure
        XXX: will be use to store avatar informations in ZODB.
        """
        pass 
    
    def getConfig(self):
        """
        get the server configuration to access to.
        @return: hostname and port
        @rtype: C{tuple} as (hostname, port)
        """
        return (self.hostname, self.port)
    
    def start(self):
        "Start the HappyChat Server"
        self.runing = True
        self.run()
        LOG('HappyChatServer', INFO, 'Starting chat server')
        
    def run(self):        
        "The main server loop "
        reactor.run(installSignalHandlers=0)   
        self.stop()
                    
    def stop(self):
        """Stop the server
        """
        reactor.stop()
        LOG('HappyChat', INFO, 'Stoping chat server')

    def isRuning(self):
        """State of the server
        @return: True is server is running
        @rtype: C{boolean}
        """
        return self.runing

    def startExternalAccess(self):
        """put the lock or wait
        """
        lock = self.lock
        lock.acquire()
        
    def stopExternalAccess(self):
        """free the lock
        """
        lock = self.lock
        lock.release()
        
    def getInfo(self):
        """
        return the server information
        as a dictionary list
        """
        self.startExternalAccess()

        info = []
        world = World()
        people = world.getAvatars()
        for avatar in people:
            avatar_info = { 'nickname' : avatar.getNickname(),
                            'address' : avatar.address }
            info.append(avatar_info)
            
        self.stopExternalAccess()

        return info
    
    def createSession(self, nickname):
        """ Create a chat session, get user information and store
        it in a temporary object before avatar creation.

        @return: the session identifier
        """
        self.startExternalAccess()
        SessionManager = HappyChatZopeSessionManager()
        session_id = SessionManager.createSession(nickname)
        self.stopExternalAccess()
        
        return session_id

                            
def run_server(id, hostname='localhost', port=1234, protocol=None):
    """The main server function : Threaded"""
    
    server = HappyServer(id, hostname, port, protocol)
    reactor.listenTCP(port, server)
    server.start()
    

def registerServer(id, hostname, port, protocol):
    """
    Register the HappyChat server for now, have to find an implementation
    to register multiple servers. You can start your own server with your
    own protocol with using this function and comment its call in this
    __init__.py
    """
    
    kw = { 'id' : id,
           'hostname' : hostname,
           'port' : port,
           'protocol' : protocol
         }
    
    HappyServer_thread = Thread(target=run_server,
                                name=id,
                                kwargs=kw)
    
    HappyServer_thread.setDaemon(1)
    HappyServer_thread.start()

    
    
    
