##############################################################################
#
# Copyright (c) 2004 Alexandre Fernandez (alex@grotte.org). All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################

import time

"""
World definition, a world is the singleton tool that manage people and room.

"""

class _World:
    """World singleton,
    this tool know all avatars, all rooms and manage them
    """
    
    instance = None

    def __init__(self):
        self.rooms = {}
        self.people = {}
        self.pending = {}
        self.connections = {}
        self.on = True

    def addRoom(self, room):
        """Add a room to the world
        @param room: Room to add
        @type room: L{Room}
        """
        self[room.id] = room

    def delRoom(self, room_id):
        """Delete room from this world by room identifier
        @param room_id: Room identifier
        @type room_id: {string}
        """
        del self[room_id]

    def __getitem__(self, room_id):
        """Get the room"""
        return self.rooms[room_id]

    def __setitem__(self, room_id, room):
        self.rooms[room_id] = room

    def born(self, avatar, room, connection):
        """Make born an avatar to this world
        @param avatar: new Avatar
        @type avatar: L{Avatar}

        @param room: The default room of the avatar
        @type room: L{Room}

        @param connection: Connection that manage the avatar
        @type connection: L{HappyChatDefaultProtocol}
        
        """
        self.people[avatar.socket] = avatar
        self.connections[connection] = avatar
        
        if room is not None:
            avatar.join(room)
        avatar.sendNickName()

    def kill(self, socket):
        """Kill the avatar by its socket
        not in use since using twisted"""
        if socket in self.pending:
            del self.pending[socket]
            socket.close()
            del socket
        else:
            avatar = self.getAvatarBySocket(socket)
            del self.people[socket]
            del avatar

    def killByConnection(self, connection):
        """Kill an avatar by the connection instance
        @param connection: Connection that manage the avatar
        @type connection: L{HappyChatDefaultProtocol}
        """
        avatar = self.getAvatarByConnection(connection)
        avatar.part()
        del self.connections[connection]
        del avatar
        
    def getAvatarBySocket(self, socket):
        return self.people[socket]

    def getAvatarByConnection(self, connection):
        """Get the avatar by the given connection
        @param connection: Connection that manage the avatar
        @type connection: L{HappyChatDefaultProtocol}
        """
        return self.connections[connection]
    
    def getAvatars(self):
        """Return all avatars of this world
        @return: Avatars
        @rtype: C{list} of L{Avatar}
        """
        return self.people.values()

    def getAvatarSockets(self):
        return self.people.keys()

    def getPendingSockets(self):
        """Get the pending connection
        @return Sockets
        @rtype: C{list} of C{socket}
        """
        return self.pending.keys()

    def getPendingInfos(self, socket):
        return self.pending[socket]

    def addPending(self, socket, addr):
        """Add a pending structure the this world
        the pending structure is the data of unborned avatar"""
        self.pending[socket] = addr
        self.ping(socket)

    def delPending(self, socket):
        del self.pending[socket]

    def ping(self, socket):
        """Send a ping command the given socket
        @param socket: destination socket to write
        @type socket: C{socket}
        """
        message = "<PING>%s</PING>\0" % time.time()
        socket.send(message)

    def nicknameExists(self, nickname):
        """
        return if nickname is in this world
        @param nickname: nickname to test
        @type nickname: C{string}

        @return: True is the nickname exists in the world
        @rtype: C{boolean}
        """
        allNickname = [avatar.getNickname().upper() for avatar in \
                       self.getAvatars()]
        return nickname.upper() in allNickname

    def getMainRoom(self):
        """
        @return: The main room
        @rtype: L{Room}
        """
        return self.rooms['main']


def World():
    if _World.instance == None:
        _World.instance = _World()
    return _World.instance
